import builtins as _builtins

from lazyutils import lazy


class Instrospection:
    """
    Introspection facilities for a transpyled Transpyler.
    """

    # Original python names
    py_constants = ['True', 'False', 'None']

    @lazy
    def py_exceptions(self):
        return [
            name for (name, value) in vars(_builtins).items()
            if isinstance(value, type) and issubclass(value, Exception)
        ]

    @lazy
    def py_types(self):
        return [
            name for (name, value) in vars(_builtins).items()
            if isinstance(value, type) and not issubclass(value, Exception)
        ]

    @lazy
    def py_functions(self):
        return [
            name for (name, value) in vars(_builtins).items()
            if name not in self.py_types and name not in self.py_exceptions
        ]

    @lazy
    def py_builtins(self):
        return self.py_types + self.py_functions

    def __init__(self, language):
        self.language = language

        # def _filtering_out(self, names):
        #     """
        #     Remove name from global _names variable
        #     """
        #
        #     for name in names:
        #         del _names[name]
        #     return names
        #
        # all_names = [name for name in dir(tugalib) if not name.startswith('_')]
        # get_namespace = {name: getattr(tugalib, name) for name in all_names}
        # _names = get_namespace.copy()
        #
        # # Classify names according to type: constants
        # constants = _filtering_out(
        #     [name for (name, value) in _names.items()
        #      if value in _py_constants or isinstance(value, (int, float))]
        # )
        #
        # # Exceptions
        # exceptions = _filtering_out(
        #     [name for (name, value) in _names.items()
        #      if isinstance(value, type) and issubclass(value, Exception)]
        # )
        #
        # # Types
        # types = _filtering_out(
        #     [name for (name, value) in _names.items() if isinstance(value, type)]
        # )
        #
        # # Functions
        # functions = _filtering_out(
        #     [name for (name, value) in _names.items() if callable(value)]
        # )
        #
        # # Modules
        # submodules = _filtering_out(
        #     [name for (name, value) in _names.items() if isinstance(value, _Module)]
        # )
        #
        # # Builtins
        # builtins = types + functions
        #
        # # Builtins dictionary
        # _original_builtins = vars(_builtins)



import functools
import re

from unidecode import unidecode

__all__ = [
    # Remove accents
    'unidecode',

    # Handle synonyms
    'synonyms', 'collect_synonyms', 'register_synonyms',

    # Other decorators
    'normalize_accented_keywords',

    # Utility functions
    'keep_spaces', 'pretty_callable', 'with_transpyler_attr',
]

HEADING_SPACE = re.compile(r'^\s*')
ENDING_SPACE = re.compile(r'\s*$')
SYNONYM_ERROR_MSG = '%s is present in get_namespace, but is also a synonym of %s'


def synonyms(*args):
    """
    Decorator that marks synonyms of a function name.
    """

    fmt_args = ['**%s**' % x for x in args]

    def decorator(func):
        data = ', '.join(fmt_args[:])
        func.__synonyms__ = args
        func.__doc__ = (func.__doc__ or '') + (
            '\n\n'
            'Notes\n'
            '-----\n\n'
            'Sinônimos: ' + data)
        return func

    return decorator


def collect_synonyms(namespace, add_unaccented=True):
    """
    Return a dictionary with all synonyms found in the given get_namespace.

    Raise a ValueError for name conflicts.
    """

    map = {}
    map.update(collect_aliases(namespace))
    if add_unaccented:
        map.update(collect_unaccented(map))
    return map


def collect_unaccented(namespace):
    map = {}
    for name, func in list(namespace.items()):
        no_accent = unidecode(name)
        if no_accent != name:
            map.setdefault(no_accent, func)
    return map


def collect_aliases(namespace):
    map = {}
    for attr, func in namespace.items():
        try:
            for alias in func.__synonyms__:
                map.setdefault(alias, func)
        except AttributeError:
            continue
    return map


def register_synonyms(global_ns, add_unaccented=True):
    """
    Register all synonyms in the given get_namespace dictionary.
    """

    D = collect_synonyms(global_ns, add_unaccented=add_unaccented)
    global_ns.update(D)


def normalize_accented_keywords(func):
    """
    Decorate function so all accented keywords are passed unaccented to the
    real implementation.
    """

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        decode = unidecode
        kwargs = {decode(k): v for (k, v) in kwargs.items()}
        return func(*args, **kwargs)

    return decorated


def keep_spaces(result, src):
    """
    Keep the same number of heading and trailing whitespace in result as
    compared to src.

    Example:
        >>> keep_spaces(' foo', 'bar\n\n')
        'foo\n\n'
    """

    if not src:
        return result.strip()

    # Find head and tail
    head = tail = ''
    if src[0].isspace():
        head = HEADING_SPACE.search(src).group()
    if src[-1].isspace():
        tail = ENDING_SPACE.search(src).group()
    return head + result.strip() + tail


def full_class_name(cls):
    """
    Return the full class name prepending module paths.
    """

    return '%s.%s' % (cls.__module__, cls.__name__)


class PrettyCallable:
    """
    Callable that whose repr() is a given message.
    """

    def __init__(self, func, name=None, doc=None, str=None,
                 autoexec=False, autoexec_message=None):
        self.__func = func
        self.__autoexec = autoexec
        self.__autoexec_message = autoexec_message
        self.__repr = str or 'please call %s()' % self.__name__

    def __call__(self, *args, **kwargs):
        return self.__func(*args, **kwargs)

    def __repr__(self):
        if self.__autoexec:
            self.__func()
            return self.__autoexec_message or ''
        return self.__repr

    def __getattr__(self, attr):
        return getattr(self.__func, attr)


def pretty_callable(str=None, **kwargs):
    """
    Decorate function to be a pretty callable.

    Example:
        >>> @pretty_callable('call exit() to finish interactive shell')
        ... def exit():
        ...     raise SystemExit
    """

    def decorator(func):
        return PrettyCallable(func, str=str, **kwargs)

    return decorator


def with_transpyler_attr(cls, transpiler):
    """
    Return a subclass with the transpyler attribute defined.
    """

    return type(cls.__name__, (cls,), {'transpyler': transpiler})



from .__meta__ import __author__, __version__
from .errors import BadSytaxError
from .builtins import Builtins
from .introspection import Instrospection
from .lexer import Lexer
from .transpyler import Transpyler



import tokenize

from lazyutils import lazy

from transpyler.errors import BadSytaxError
from transpyler.token import Token, displace_tokens, token_find
from transpyler.utils import keep_spaces


class Lexer:
    """
    Lexer is responsible for converting a string of source code into a sequence
    of tokens.
    """

    @lazy
    def invalid_tokens(self):
        try:
            return dict(self.language.invalid_tokens)
        except (AttributeError, TypeError):
            return {}

    @lazy
    def translations(self):
        try:
            return dict(self.language.translations)
        except (AttributeError, TypeError):
            return {}

    @lazy
    def sequence_translations(self):
        return {tuple(k): v
                for k, v in self.translations.items()
                if not isinstance(k, str)}

    @lazy
    def single_translations(self):
        return {k: v
                for k, v in self.translations.items()
                if isinstance(k, str)}

    def __init__(self, language):
        self.language = language

    def transpile(self, src):
        """
        Transpile source code to Python.
        """

        # Avoid problems with empty token streams
        if not src or src.isspace():
            return src

        # Convert and process...
        else:
            src_formatted = src

            if not src_formatted.endswith('\n'):
                src_formatted += '\n'

            tokens = self.tokenize(src_formatted)
            transpiled_tokens = self.transpile_tokens(tokens)
            result = self.untokenize(transpiled_tokens)
            return keep_spaces(result, src)

    def tokenize(self, src):
        """
        Convert source string to a list of tokens.

        Args:
            src (str): a string of source code
        """

        current_string = src

        # FIXME: Not reaching here!
        # if not current_string.endswith('\n'):
        #    current_string += '\n'

        def iterlines():
            nonlocal current_string

            if current_string:
                line, sep, current_string = current_string.partition('\n')
                return line + sep
            else:
                raise StopIteration

        tokens = list(tokenize.generate_tokens(iterlines))
        tokens = list(map(Token, tokens))
        return tokens

    def untokenize(self, tokens):
        """
        Convert list of tokens to a string.
        """

        return tokenize.untokenize([tk.to_token_info() for tk in tokens])

    def transpile_tokens(self, tokens):
        """
        Transpile a sequence of Token objects to their corresponding Python
        tokens.
        """

        self.detect_error_sequences(tokens, self.invalid_tokens)
        tokens = self.replace_sequences(tokens, self.sequence_translations)
        tokens = self.replace_translations(tokens, self.single_translations)
        return tokens

    def detect_error_sequences(self, tokens, error_dict):
        """
        Raises a BadSyntaxError if list of tokens contains any sub-sequence in
        the given invalid_tokens.

        Args:
            tokens: List of tokens
            error_dict: A dictionary of {sequence: error_message}
        """

        error_sequences = list(error_dict)
        iterator = token_find(tokens, error_sequences)
        while True:
            try:
                idx, match, start, end = next(iterator)
                msg = error_dict(match)
                match = ' '.join(match)
                raise BadSytaxError(msg, from_token=tokens[idx])
            except StopIteration:
                break

    def replace_sequences(self, tokens, mapping):
        """
        Replace all sequences of tokens in the mapping by the corresponding
        token in the RHS.

        Args:
            tokens: list of tokens.
            mapping: a mapping from token sequences to their corresponding
                replacement (e.g.: {('para', 'cada'): 'for'}).

        Returns:
            A new list of tokens with replacements.
        """
        tokens = list(tokens)
        iterator = token_find(tokens, mapping)
        while True:
            try:
                idx, match, start, end = next(iterator)
                tokens[idx] = Token(mapping[match], start=start)
                del tokens[idx + 1: idx + len(match)]

                linediff, col = tokens[idx].end - end
                if linediff == 0:
                    displace_tokens(tokens[idx + 1:], col)
            except StopIteration:
                break
        return tokens

    def replace_translations(self, tokens, mapping):
        """
        Replace all tokens by the corresponding values in the RHS.

        Args:
            tokens: list of tokens.
            mapping: a mapping from token sequences to their corresponding
                replacement (e.g.: {'enquanto': 'while'}).

        Returns:
            A new list of tokens with replacements.
        """

        tokens = list(tokens)
        for i, tk in enumerate(tokens):
            new = mapping.get(tk.string, tk)
            if new is not tk:
                new = Token(new, start=tk.start)
                tokens[i] = new

                # Align tokens
                linediff, coldiff = new.end - tk.end
                assert linediff == 0
                if coldiff:
                    displace_tokens(tokens[i + 1:], coldiff)
        return tokens



import pytest

from transpyler import Transpyler


@pytest.fixture
def python():
    """
    A new transpyler that makes no changes in comparison with the default
    interpreter.
    """

    class Python(Transpyler):
        pass

    return Python()


@pytest.fixture
def pybr():
    class PyBr(Transpyler):
        translations = {
            'para': 'for',
            'em': 'in',
            ('para', 'cada'): 'for',
            ('faça', ':'): ':',
        }
        builtin_modules = ['math']

    return PyBr()


@pytest.fixture
def pybr_code():
    return '''
x, y = 1, 1
para cada i em [1, 2, 3, 4, 5] faça:
    x, y = y, x + y
'''



from transpyler.utils import *


def test_keep_spaces():
    f = keep_spaces
    assert f(' foo ', '') == 'foo'
    assert f('foo', 'bar') == 'foo'
    assert f(' \nfoo', 'bar') == 'foo'
    assert f(' \nfoo\n ', 'bar') == 'foo'
    assert f('foo', ' \nbar') == ' \nfoo'
    assert f('\n foo\n ', ' \nbar') == ' \nfoo'
    assert f('\n foo', ' \nbar') == ' \nfoo'
    assert f('foo\n', ' \nbar') == ' \nfoo'
    assert f('foo', ' \nbar\n ') == ' \nfoo\n '
    assert f('\n foo', ' \nbar\n ') == ' \nfoo\n '
    assert f('\n foo\n ', ' \nbar\n ') == ' \nfoo\n '
    assert f('foo\n ', ' \nbar\n ') == ' \nfoo\n '
    assert f('foo\n   bar', 'ham\n   spam') == 'foo\n   bar'


def test_synonyms():
    @synonyms('foobar', 'bar')
    def foo(x, y):
        pass

    sym_ns = collect_synonyms({'foo': foo})
    assert sym_ns['foobar'] is foo
    assert sym_ns['bar'] is foo
    assert 'foo' not in sym_ns


def test_synonyms_unaccented():
    @synonyms('fôo', 'bár')
    def fôobar(x, y):
        pass

    foo = fôobar

    sym_ns = collect_synonyms({'fôobar': foo}, add_unaccented=True)
    assert sym_ns == {
        'foo': foo,
        'bar': foo,
        'fôo': foo,
        'bár': foo,
    }


def test_normalized_kwargs():
    @normalize_accented_keywords
    def foo(bar, baz):
        return bar + 2 * baz

    assert foo(1, 2) == 5
    assert foo(bar=1, baz=2) == 5
    assert foo(bár=1, báz=2) == 5


def test_register_synonyms():
    def foo(): pass

    @synonyms('spam', 'spám')
    def eggs():
        pass

    ns = {'foo': foo, 'eggs': eggs}
    register_synonyms(ns)

    assert ns == {
        'foo': foo, 'eggs': eggs, 'spam': eggs, 'spám': eggs
    }


def test_pretty_function():
    @pretty_callable('type foo() to call it')
    def foo():
        return 42

    assert repr(foo) == 'type foo() to call it'
    assert foo() == 42
    assert foo.__name__ == 'foo'


def test_with_transpyler_class():
    class Foo:
        pass

    Bar = with_transpyler_attr(Foo, 'transpyler')
    assert Bar.transpyler == 'transpyler'



"""
Makes ipytuga kernel discoverable by Jupyter
"""

import os
import sys

assets_dir = os.path.join(os.path.dirname(__file__), 'assets')


def setup_assets(transpyler, user=False, ipython_dir=None):
    from ipykernel.kernelspec import KernelSpecManager
    from jupyter_client.kernelspec import install_kernel_spec

    if ipython_dir is None:
        install = install_kernel_spec
    else:
        spec = KernelSpecManager(ipython_dir=ipython_dir)
        install = spec.install_kernel_spec

    with transpyler.info.prepare_assets() as path:
        install(path, transpyler.name, replace=True, user=user)


if __name__ == '__main__':
    setup_assets(user='--user' in sys.argv)



from transpyler.jupyter import start_shell
from ..transpyler import get_default_transpyler


# Executes using the global default transpyler instance
if __name__ == '__main__':
    transpyler = get_default_transpyler()
    start_shell(transpyler)



import json
import os
import tempfile
from contextlib import contextmanager

from transpyler.utils import full_class_name


class LanguageInfo:

    def __init__(self, transpyler):
        self.transpyler = transpyler

    def get_argv(self):
        """
        Return the list of arguments to be passed to initialize the kernel.
        """
        return [
            "python3", "-m", "transpyler.jupyter.kernel",
            "-f", "{connection_file}",
            "--type", full_class_name(type(self.transpyler)),
        ]

    def get_pygments_lexer(self):
        return 'ipython3'

    def get_kernel_spec(self):
        """
        Return the kernel spec as a JSON-like data.
        """

        transpyler = self.transpyler
        return {
            "argv": self.get_argv(),
            "display_name": transpyler.display_name,
            "language": transpyler.name,
            "codemirror_mode": transpyler.codemirror_mode,
            "language_info": {
                "name": transpyler.display_name,
                "codemirror_mode": {
                    "name": transpyler.codemirror_mode,
                },
                "mimetype": transpyler.mimetype,
                "pygments_lexer": self.get_pygments_lexer(),
            },
            "help_links": [
                {
                    "text": "Documentação do Pytuguês",
                    "link": "http://pytuga.readthedocs.io/pt/latest/"
                },
                {
                    "text": "Github",
                    "link": "http://github.com/transpyler/transpyler/"
                }
            ]
        }

    def get_assets(self):
        """
        Register a list of all assets available for the given kernel.
        """

        basedir = os.path.dirname(__file__)
        assets_dir = os.path.join(basedir, 'jupyter', 'assets')
        filenames = os.listdir(assets_dir)
        return [os.path.join(assets_dir, name) for name in filenames]

    @contextmanager
    def prepare_assets(self):
        """
        Creates a temporary directory with all assets.
        """

        temp = tempfile.mkdtemp()
        try:
            for asset in self.get_assets():
                name = os.path.split(asset)[-1]
                with open(asset, 'rb') as input:
                    with open(os.path.join(temp, name), 'wb') as output:
                        output.write(input.read())
            with open(os.path.join(temp, 'kernel.json'), 'w') as F:
                dump = json.dumps(self.get_kernel_spec(), indent=2)
                F.write(dump)
            yield temp
        finally:
            for file in os.listdir(temp):
                os.unlink(os.path.join(temp, file))
            os.rmdir(temp)



import importlib
from ast import PyCF_ONLY_AST

from IPython.core.compilerop import CachingCompiler
from IPython.utils import py3compat
from ipykernel.ipkernel import IPythonKernel
from ipykernel.zmqshell import ZMQInteractiveShell
from lazyutils import lazy
from traitlets import Type

from transpyler.utils import with_transpyler_attr
from transpyler.transpyler import get_transpyler_from_name


class TranspylerShell(ZMQInteractiveShell):
    """
    A IPython based shell for transpyled languages.

    A shell is used by the kernel to interact with code.
    """

    @lazy
    def transpyler(self):
        return self.parent.transpyler

    def init_user_ns(self):
        """
        Additional symbols for the shell environment.
        """

        super().init_user_ns()
        ns = self.user_ns
        self.transpyler.update_user_ns(ns)

    def ex(self, cmd):
        return super().ex(self.transpyler.transpile(cmd))

    def ev(self, cmd):
        return super().ev(self.transpyler.transpile(cmd))


class TranspylerKernel(IPythonKernel):
    """
    A meta kernel based backend to use Transpyled languages in Jupyter/iPython.
    """

    transpyler = None

    @lazy
    def implementation(self):
        return 'i' + self.transpyler.name

    @lazy
    def implementation_version(self):
        return self.transpyler.version

    @lazy
    def language(self):
        return self.transpyler.name

    @lazy
    def language_version(self):
        return self.transpyler.language_version

    @lazy
    def banner(self):
        return self.transpyler.get_console_banner()

    @lazy
    def language_info(self):
        transpyler = self.transpyler
        return {
            'mimetype': transpyler.mimetype,
            'file_extension': transpyler.file_extension,
            'codemirror_mode': {
                "version": 3,
                "name": "ipython"
            },
            'pygments_lexer': 'python',
        }

    shell_class = Type(TranspylerShell)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.transpyler is None:
            raise ValueError('transpyler was not defined')
        monkey_patch(self.transpyler)
        self.transpyler.init()

    def do_execute(self, code, *args, **kwargs):
        code = self.transpyler.transpile(code)
        return super().do_execute(code, *args, **kwargs)

    def do_is_complete(self, code):
        return super().do_is_complete(self.transpyler.transpile(code))


def start_kernel(transpyler):
    """
    Start Pytuga Jupyter kernel.
    """

    from ipykernel.kernelapp import IPKernelApp

    kernel_class = with_transpyler_attr(TranspylerKernel, transpyler)
    IPKernelApp.launch_instance(kernel_class=kernel_class)


# We mokey-patch a few modules and functions to replace Python's compile
# function with pytuga's version. This is potentially fragile, but it seems to
# work well so far.
def monkey_patch(transpyler):
    def ast_parse(self, source, filename='<unknown>', symbol='exec'):
        flags = self.flags | PyCF_ONLY_AST
        return transpyler.compile(source, filename, symbol, flags, 1)

    CachingCompiler.ast_parse = ast_parse
    py3compat.compile = transpyler.compile


# Find correct transpyler instance from argv and execute start_kernel
if __name__ == '__main__':
    import sys

    idx = sys.argv.index('--type')
    sys.argv.pop(idx)
    transpyler_class = sys.argv.pop(idx)
    path, _, name = transpyler_class.rpartition('.')
    transpyler_class = getattr(importlib.import_module(path), name)
    start_kernel(get_transpyler_from_name(transpyler_class))



import re

from pygments import unistring as uni
from pygments.lexer import bygroups, default, words
from pygments.lexers.python import Python3Lexer, PythonTracebackLexer
from pygments.token import Text, Operator, Keyword, Name, String, Number
from pygments.util import shebang_matches

from pytuga import core as tuga_introspect
from pytuga import keyword

KEYWORDS = keyword.kwlist
CONSTANTS = keyword.constants
EXCEPTIONS = tuga_introspect.py_exceptions + tuga_introspect.exceptions
BUILTINS = tuga_introspect.py_builtins + tuga_introspect.builtins
NAME = 'Transpyled'
ALIASES = ['transpyled']
FILENAMES = ['pytp']
MIMETYPES = ['text/x-transpyled', 'application/x-transpyled']


class TranspyledLexer(Python3Lexer):
    """
    A lexer for transpyler based languages.

    This is just the standard Python
    """

    name = NAME
    aliases = ALIASES
    filenames = FILENAMES
    mimetypes = MIMETYPES

    flags = re.MULTILINE | re.UNICODE

    uni_name = "[%s][%s]*" % (uni.xid_start, uni.xid_continue)

    tokens = Python3Lexer.tokens.copy()  # @UndefinedVariable

    tokens['keywords'] = [
        (words(KEYWORDS, suffix=r'\b'), Keyword),
        (words(CONSTANTS, suffix=r'\b'), Keyword.Constant),
    ]
    tokens['builtins'] = [
        (words(BUILTINS, prefix=r'(?<!\.)', suffix=r'\b'), Name.Builtin),
        (r'(?<!\.)(self|Ellipsis|NotImplemented)\b', Name.Builtin.Pseudo),
        (words(EXCEPTIONS, prefix=r'(?<!\.)', suffix=r'\b'),
         Name.Exception),
    ]
    tokens['numbers'] = [
        (r'(\d+\.\d*|\d*\.\d+)([eE][+-]?[0-9]+)?', Number.Float),
        (r'0[oO][0-7]+', Number.Oct),
        (r'0[bB][01]+', Number.Bin),
        (r'0[xX][a-fA-F0-9]+', Number.Hex),
        (r'\d+', Number.Integer)
    ]
    tokens['backtick'] = []
    tokens['name'] = [
        (r'@\w+', Name.Decorator),
        (uni_name, Name),
    ]
    tokens['funcname'] = [
        (uni_name, Name.Function, '#pop')
    ]
    tokens['classname'] = [
        (uni_name, Name.Class, '#pop')
    ]
    tokens['import'] = [
        (r'(\s+)(as)(\s+)', bygroups(Text, Keyword, Text)),
        (r'\.', Name.Namespace),
        (uni_name, Name.Namespace),
        (r'(\s*)(,)(\s*)', bygroups(Text, Operator, Text)),
        default('#pop')  # all else: go back
    ]
    tokens['fromimport'] = [
        (r'(\s+)(import)\b', bygroups(Text, Keyword), '#pop'),
        (r'\.', Name.Namespace),
        (uni_name, Name.Namespace),
        default('#pop'),
    ]
    tokens['strings'] = [
        # the old style '%s' % (...) string formatting (still valid in Py3)
        (r'%(\(\w+\))?[-#0 +]*([0-9]+|[*])?(\.([0-9]+|[*]))?'
         '[hlL]?[diouxXeEfFgGcrs%]', String.Interpol),
        # the new style '{}'.format(...) string formatting
        (r'\{'
         '((\w+)((\.\w+)|(\[[^\]]+\]))*)?'  # field name
         '(\![sra])?'  # conversion
         '(\:(.?[<>=\^])?[-+ ]?#?0?(\d+)?,?(\.\d+)?[bcdeEfFgGnosxX%]?)?'
         '\}', String.Interpol),
        # backslashes, quotes and formatting signs must be parsed one at a time
        (r'[^\\\'"%\{\n]+', String),
        (r'[\'"\\]', String),
        # unhandled string formatting sign
        (r'%|(\{{1,2})', String)
        # newlines are an error (use "nl" state)
    ]

    def analyse_text(text):  # @NoSelf
        return shebang_matches(text, r'pythonw?3(\.\d)?')


class PytugaTracebackLexer(PythonTracebackLexer):
    pass



#
# Test common transpyler examples for python transpyler
#
import pytest


def test_language_transpile_code(python):
    src = 'print("hello world")'
    assert python.transpile(src) == src


def test_language_eval_code(python):
    assert python.eval('1 + 1', {}) == 2


def test_language_exec_code(python):
    D = {}
    python.exec('x = 42', D)
    assert D['x'] == 42


#
# Test common transpyler constructs for PyBr transpyler
#
def test_pybr_eval(pybr):
    assert pybr.eval('1 em (1, 2)', {}) is True


def test_pybr_code_transpile(pybr, pybr_code):
    assert pybr.transpile(pybr_code) == (
        '\n'
        'x, y = 1, 1\n'
        'for i in [1, 2, 3, 4, 5] :\n'
        '    x, y = y, x + y\n'
    )


def test_pybr_code(pybr, pybr_code):
    D = {}
    pybr.exec(pybr_code, D)
    assert D['x'] == 8
    assert D['y'] == 13


def test_pybr_builtins(pybr):
    ns = pybr.get_builtins_namespace()
    assert 'cos' in ns


def test_pybr_exec_with_builtins(pybr):
    assert pybr.eval('cos(0)', {}) == 1


def test_pybr_exec_with_temp_builtins(pybr):
    D = {}
    assert pybr.eval('cos(0)', D, builtins=True) == 1
    assert 'cos' not in D


#
# Tests error code paths for PyBr
#
def test_empty_code_string(pybr):
    assert pybr.transpile('') == ''


def test_invalid_keywords_tokens(pybr):
    with pytest.raises(SyntaxError):
        pybr.exec('Retornar 1', {})



import os

from jupyter_client import KernelManager
from jupyter_console.app import ZMQTerminalIPythonApp
from jupyter_console.ptshell import ZMQTerminalInteractiveShell


class TranspylerKernelManager(KernelManager):
    transpyler = None

    @property
    def kernel_name(self):
        return self.transpyler.name

    @property
    def kernel_script_path(self):
        return os.path.join(os.path.dirname(__file__), 'kernel.py')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.transpyler is None:
            raise ValueError('did not set transpyler instance')


class ZMQTerminalTranspylerApp(ZMQTerminalIPythonApp):
    """
    A Transpyler terminal app.
    """

    transpyler = None
    name = 'transpyler'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.transpyler is None:
            raise ValueError('did not set transpyler instance')


# ZMQTerminalInteractiveShell class. It is a singleton, hence we don't expect
# problems.
#
# These hacks customize the messages shown during shell initialization.
def monkey_patch():
    def show_banner(self):
        print(self.kernel_info.get('banner', ''))

    ZMQTerminalInteractiveShell.show_banner = show_banner


monkey_patch()



class BadSytaxError(SyntaxError):
    """
    Base class for transpyler syntax errors.
    """

    @property
    def msg(self):
        return self.args[0]

    @property
    def lineno(self):
        return self.args[1]

    @property
    def pos(self):
        return self.args[2]

    def __init__(self, msg, lineno=None, pos=None, from_token=None):
        if from_token:
            lineno = from_token.start.lineno
            pos = from_token.start.pos
        args = msg, lineno, pos
        super().__init__(args)

    def __str__(self):
        if self.lineno:
            return 'at line %s, %s: %s' % (self.lineno, self.pos, self.msg)
        return self.msg



import code
import sys
import traceback

from lazyutils import lazy


class TranspylerConsole(code.InteractiveConsole):
    """
    Very simple console for Transpyled languages.
    """

    transpyler_class = None

    @lazy
    def transpyler(self):
        return self.transpyler_class()

    def __init__(self, locals=None, filename='<console>', transpyler=None):
        if transpyler is not None:
            self.transpyler = transpyler
        locals = locals if locals is not None else {}
        locals.update(self.transpyler.get_builtins_namespace())
        super().__init__(locals, filename)

    def runsource(self, source, filename="<input>", symbol="single"):
        try:
            source = self.transpyler.transpile(source)
            if source.endswith('\n'):
                source = source[:-1]
            code = self.compile(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            # Case 1
            print(source)
            self.showsyntaxerror(filename)
            return False

        if code is None:
            # Case 2
            return True

        # Case 3
        self.runcode(code)
        return False

    def runcode(self, code):
        if isinstance(code, str):
            code = self.transpyler.transpile(code)
        super().runcode(code)

    def showsyntaxerror(self, filename=None):
        type, value, tb = sys.exc_info()
        sys.last_type = type
        sys.last_value = value
        sys.last_traceback = tb

        if filename and type is SyntaxError:
            # Work hard to stuff the correct filename in the exception
            try:
                msg, (dummy_filename, lineno, offset, dummy_line) = value.args
            except ValueError:
                # Not the format we expect; leave it alone
                pass
            else:
                # Stuff in the right filename and line
                print(self.buffer, lineno)
                line = dummy_line
                value = SyntaxError(msg, (filename, lineno, offset, line))
                sys.last_value = value
        if sys.excepthook is sys.__excepthook__:
            lines = traceback.format_exception_only(type, value)
            self.write(''.join(lines))
        else:
            # If someone has set sys.excepthook, we let that take precedence
            # over self.write
            sys.excepthook(type, value, tb)

    def interact(self, banner=None, exitmsg=None):
        """
        Starts the console mainloop.
        """

        self.transpyler.init()
        super().interact(banner, exitmsg)


def run_console(transpyler):
    """
    Run the main console.
    """

    console = TranspylerConsole(transpyler=transpyler)
    banner = transpyler.get_console_banner()
    console.interact(banner)


if __name__ == '__main__':
    run_console()



import tokenize
from token import OP, NAME, NUMBER
from tokenize import TokenInfo, EXACT_TOKEN_TYPES

TOKEN_TYPE_NAME = {tt: attr for (attr, tt) in vars(tokenize).items()
                   if attr.isupper() and isinstance(tt, int)}


class Token:
    """
    Mutable token object.
    """

    @classmethod
    def from_strings(cls, start, *strings):
        """
        Return a list of strings starting at the given starting point and
        return the corresponding strings with the correct start/end positions
        """

        tk_list = []
        start = TokenPosition(start)
        is_fragile = False
        for string in strings:
            if is_fragile and string.isidentifier():
                start += (0, 1)
            tok = Token(string, start=start)
            start = tok.end
            is_fragile = string.isidentifier()
            tk_list.append(tok)
        return tk_list

    def __init__(self, data, type=None, start=None, end=None, line=None,  # noqa: C901
                 abstract=False):

        # Start from TokenInfo object
        if isinstance(data, TokenInfo):
            assert all(x is None for x in [type, start, end, line])
            type, data, start, end, line = data

        elif isinstance(data, Token):
            assert all(x is None for x in [type, start, end, line])
            data, type, start, end, line = data

        elif isinstance(data, int):
            type = data
            data = None

        if start is not None:
            start = TokenPosition(start)
            if end is None:
                if '\n' not in data:
                    end = start + (0, len(data))
                else:
                    lineno = end.lineno + end.count('\n')
                    col = len(data.rpartition('\n')[-1])
                    end = TokenPosition(lineno, col)
        if end is not None:
            end = TokenPosition(end)
        if type is None:
            # We are not using exact token types: the tokenizer converts all
            # of them to OP.
            if data in EXACT_TOKEN_TYPES:
                type = OP
            else:
                if data.isidentifier():
                    type = NAME
                elif data.isdigit():
                    type = NUMBER
                else:
                    raise TypeError('could not recognize token: %r' % data)

        self.string = data
        self.type = type
        self.start = start
        self.end = end
        self.line = line
        if not abstract:
            if start is None or end is None:
                raise ValueError(
                    'could not define start/end of concrete token')
        else:
            if start is not None or end is not None:
                raise ValueError('cannot define start/end positions of '
                                 'abstract token')

    def __eq__(self, other):  # noqa: C901
        if isinstance(other, Token):
            if self.string is None or other.string is None:
                return self.type == other.type

            if self.string != other.string or self.type != other.type:
                return False

            for (x, y) in [(self.start, other.start),
                           (self.end, other.end),
                           (self.line, other.line)]:
                if x != y and None not in (x, y):
                    return False
            return True
        elif isinstance(other, (str, TokenInfo)):
            return self == Token(other)
        elif isinstance(other, (tuple, list)):
            if len(other) == 2:
                return self.string == other[0] and self.type == other[1]
            return self == Token(*other)
        else:
            return NotImplemented

    def __str__(self):
        return 'Token(%r, %s, %r, %r, %r)' % (
            self.string, TOKEN_TYPE_NAME[self.type],
            self.start, self.end, self.line)

    def __repr__(self):
        tname = TOKEN_TYPE_NAME[self.type]
        if self.string.isspace():
            return tname
        else:
            return '%s(%r)' % (tname, self.string)

    def __len__(self):
        return 5

    def __getitem__(self, idx):
        if idx == 0:
            return self.string
        elif idx == 1:
            return self.type
        elif idx == 2:
            return self.start
        elif idx == 3:
            return self.end
        elif idx == 4:
            return self.line
        elif -5 <= idx < 0:
            return self[5 + idx]
        else:
            raise IndexError(idx)

    def __iter__(self):
        yield from (self[i] for i in range(5))

    def to_token_info(self):
        """
        Convert to TokenInfo object used by Python's tokenizer.
        """

        return TokenInfo(
            self.type, self.string, self.start, self.end, self.line
        )

    def displace(self, cols):
        """
        Displace token in line by cols columns to the right.
        """

        self.start += (0, cols)
        if self.end.lineno == self.start.lineno:
            self.end += (0, cols)


class TokenPosition(tuple):
    """
    Represent the start or end position of a token and accept some basic
    arithmetic operations
    """

    def __new__(cls, x, y=None):
        if y is None:
            x, y = x
        return tuple.__new__(cls, [x, y])

    def __init__(self, x, y=None):
        super().__init__()

    def __add__(self, other):
        x, y = self
        a, b = other
        return TokenPosition(x + a, y + b)

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        x, y = self
        a, b = other
        return TokenPosition(x - a, y - b)

    def __rsub__(self, other):
        x, y = self
        a, b = other
        return TokenPosition(a - x, b - y)

    @property
    def lineno(self):
        return self[0]

    @property
    def col(self):
        return self[1]


def displace_tokens(tokens, cols):
    """
    Displace all tokens in list which are in the same line as the the first
    token by the given number of columns
    """

    if not tokens or cols == 0:
        return
    lineno = tokens[0].start.lineno

    for token in tokens:
        if token.start.lineno == lineno:
            token.displace(cols)
        else:
            break


def insert_tokens_at(tokens, idx, new_tokens, end=None):
    """
    Insert new_tokens at tokens list at the given idx
    """

    if end is not None:
        linediff, col = new_tokens[-1].end - end
        if linediff == 0 and end.lineno == tokens[idx].start.lineno:
            displace_tokens(tokens[idx:], col)

    for tk in new_tokens:
        tokens.insert(idx, tk)
        idx += 1


def token_find(tokens, matches, start=0):
    """
    Iterate over list of tokens yielding (index, match, start, end) for
    each match in the token stream. The `matches` attribute must be a sequence
    of token sequences.
    """

    matches = list(matches)
    tk_matches = \
        [tuple(Token(tk, abstract=True) for tk in seq) for seq in matches]
    tk_matches = \
        [tuple((tk.string, tk.type) for tk in seq) for seq in tk_matches]
    tk_idx = start

    while tk_idx < len(tokens):
        for match_idx, tkmatch in enumerate(tk_matches):
            if tk_idx + len(tkmatch) > len(tokens):
                continue

            if all(tokens[tk_idx + k] == tk for (k, tk) in enumerate(tkmatch)):
                matchsize = len(tkmatch)

                # Yield value and try to receive commands from sender
                start = tokens[tk_idx].start
                end = tokens[tk_idx + matchsize - 1].end
                yield (tk_idx, matches[match_idx], start, end)
                tk_idx += 1
        tk_idx += 1
    return



import contextlib
import builtins as _builtins
import importlib

from lazyutils import lazy


class Builtins:
    """
    Manage the builtin functions for the new transpyler.
    """

    @lazy
    def modules(self):
        try:
            return list(self.transpyler.builtin_modules)
        except (AttributeError, TypeError):
            return []

    @lazy
    def _namespace(self):
        ns = {}
        for mod in self.modules:
            mod = importlib.import_module(mod)
            for name in dir(mod):
                if name.startswith('_') or name.isupper():
                    continue
                ns[name] = getattr(mod, name)
        return ns

    def __init__(self, transpyler=None):
        self.transpyler = transpyler

    def get_namespace(self):
        """
        Return a dictionary with all public functions.
        """

        return dict(self._namespace)

    def update(self, *args, **kwargs):
        """
        Update builtins with new functions.
        """

        ns = self._namespace
        ns.update(*args, **kwargs)

    def load_as_builtins(self):
        """
        Load all registered builtin functions as Python's builtins.
        """

        for k, v in self.get_namespace().items():
            setattr(_builtins, k, v)

    @contextlib.contextmanager
    def restore_after(self):
        """
        Apply current builtins to Python's builtins module and restore
        state afterwards.
        """

        old_builtins = vars(_builtins)
        new_builtins = self.get_namespace()

        for k, v in new_builtins.items():
            setattr(_builtins, k, v)

        try:
            yield
        finally:
            for k in dir(_builtins):
                if k in new_builtins:
                    if k in old_builtins:
                        setattr(_builtins, k, old_builtins[k])
                    else:
                        delattr(_builtins, k)



import inspect

from textblob import TextBlob


def translate_functions(seq, to_lang, from_lang='en'):
    """
    Use Google Translate to translate names, parameters and docstrings for
    the given list of functions.
    """

    return [translate_function(func, to_lang, from_lang='en') for func in seq]


def translate_function(function, to_lang, from_lang='en'):
    """
    Translate a single function using Google Translate. The resulting source
    code just calls the original function and assumes it lives in the global
    get_namespace.
    """

    name = TextBlob(function.__name__)
    args = get_args(function)
    docs = TextBlob(function.__doc__)
    lines = []

    # Format arguments declaration
    args_decl = []
    for (arg_name, default) in args:
        if default is None:
            args_decl.append(arg_name)
        else:
            args_decl.append('%s=%s' % (arg_name, default))
    args_decl = ', '.join(args_decl)

    # Translate function name and make declaration.
    translated_name = str(name.translate(from_lang, to_lang))
    translated_name = translated_name[0].lower() + translated_name[1:]
    lines.append('def %s(%s):' % (translated_name, args_decl))

    # Translate docstring, indent and add it to lines
    translated_docs = docs.translate(from_lang, to_lang)
    lines.append('    """')
    for line in str(translated_docs).splitlines():
        lines.append('    ' + line)
    lines.append('    """')

    # Call original function
    args_call = []
    has_stararg = False
    for arg_name, default in args:
        if has_stararg and not arg_name.startswith('*'):
            args_call.append('%s=%s' % (arg_name, arg_name))
        else:
            args_call.append(arg_name)
            if arg_name.startswith('*'):
                has_stararg = True
    args_call = ', '.join(args_call)
    lines.append('    return %s(%s)' % (name, args_call))

    # Render source code
    return '\n'.join(lines)


def get_args(function):
    """
    Return a list of named arguments (name, default) tuples. If it has no
    default, return (name, None). Variational positional and keyword arguments
    are also returned, as ('*args', None) and ('**kwargs, None).
    """

    try:
        spec = inspect.getfullargspec(function)
        return get_args_from_spec(spec)
    except TypeError:
        return get_args_from_docstring(function)


def get_args_from_spec(spec):
    args = []

    # Save default args
    diff = len(spec.args) - len(spec.defaults)
    defaults = (None,) * diff + spec.defaults
    defaults = [x if x is None else repr(x) for x in defaults]
    args.extend(zip(spec.args, defaults))

    # Add varags
    if spec.varargs:
        args.append(('*%s' % spec.varargs, None))
    elif spec.kwonlyargs:
        args.append('*', None)

    # Add keyword only
    if spec.kwonlyargs:
        defaults = [repr(spec.kwonlydefaults[k]) for k in spec.kwonlyargs]
        args.extend(zip(spec.kwonlyargs, defaults))

    # Add varkw
    if spec.varkw:
        args.append(('**%s' % spec.varkw, None))
    return args


def get_args_from_docstring(function):
    signature = function.__doc__.partition('\n')[0]
    name = function.__name__
    if not signature.startswith(name):
        raise TypeError('could not determine signature of builtin.')
    signature = signature.partition('(')[2].partition(')')[0]
    signature = signature.replace('[', '').replace(']', '')
    return [(x.strip(), None) for x in signature.split(',')]



import ctypes

from transpyler.utils import collect_synonyms

cpython = ctypes.pythonapi


class SlotsProxy(ctypes.Structure):
    _fields_ = [
        ('ob_refcnt', ctypes.c_ssize_t),
        ('ob_type', ctypes.py_object),
        ('dict', ctypes.py_object),
    ]


def add_attr_curse(obj, attr, value):
    """
    Adds a new attribute to builtin object cls.
    """

    target = obj.__dict__
    proxy_dict = SlotsProxy.from_address(id(target))
    proxy_dict.dict[attr] = value


def apply_curse_to_class(tt, curse):
    """
    Monkey patch the builtin type tt to have all methods form the given curse
    class. Name and docstring are also updated.
    """

    namespace = {k: v for (k, v) in vars(curse).items() if not k[0] == '_'}
    functions = collect_synonyms(namespace)
    for name, func in functions.items():
        if hasattr(tt, name):
            tname = tt.__name__
            raise ValueError('%s already has a %s() method' % (tname, name))
        add_attr_curse(tt, name, func)

    # Apply docstring (not working)
    add_attr_curse(tt, '__doc__', curse.__doc__)


def apply_curses(curse_map):
    """
    Apply all curses defined in the curse map.

    Args:
        curse_map:
            A dictionary of (type, curse) pairs.
    """

    for tt, curse in curse_map.items():
        apply_curse_to_class(tt, curse)


class _object:
    """
    Generic type we use to inspect the location of the generic C-level
    repr and str functions.

    All methods supported must provide a placeholder implementation here.
    """

    __repr__ = __str__ = lambda self: ''


def _repr_offset():
    """How many bytes after id(type) can we find the tp_repr pointer?"""

    return 11 * ctypes.sizeof(ctypes.c_ssize_t)


def _str_offset():
    """How many bytes after id(type) can we find the tp_str pointer?"""

    # str is just 6 places after repr in the type specification
    return _repr_offset() + 6 * ctypes.sizeof(ctypes.c_ssize_t)


def _assure_generic_c_level_function(tt, offset):
    """
    Makes sure that the given type tt uses the generic c-level function
    for some of the magic methods such as repr(), str(), etc.
    """

    ref_from_address = ctypes.c_ssize_t.from_address
    tp_func_object = ref_from_address(id(_object) + offset)
    tp_func_cursed = ref_from_address(id(tt) + offset)
    tp_func_cursed.value = tp_func_object.value


def curse_bool_repr(true='True', false='False'):
    """
    Change repr of True/False to the given input strings.
    """

    def __repr__(self):
        if self:
            return true
        else:
            return false

    add_attr_curse(bool, '__repr__', __repr__)
    add_attr_curse(bool, '__str__', __repr__)
    _assure_generic_c_level_function(bool, _str_offset())
    _assure_generic_c_level_function(bool, _repr_offset())


def curse_none_repr(name):
    """
    Change repr of None to the given string.
    """

    name = name or 'None'

    def __repr__():
        return name

    add_attr_curse(type(None), '__repr__', __repr__)
    add_attr_curse(type(None), '__str__', __repr__)
    _assure_generic_c_level_function(type(None), _str_offset())
    _assure_generic_c_level_function(type(None), _repr_offset())



import builtins as _builtins
import codeop

from lazyutils import lazy

from .builtins import Builtins
from .language_info import LanguageInfo
from .lexer import Lexer

# Save useful builtin functions
_compile = _builtins.compile
_exec = _builtins.exec
_eval = _builtins.eval
_input = _builtins.input
_print = _builtins.print

INSTANCES_FOR_NAME = {}


class Transpyler:
    """
    Base class for a new transpyled transpyler.

    Very simple Python variations can be created by subclassing Transpyler::

        class PyBr(Transpyler):
            translations = {
                'para': 'for',            # single token translations
                'em': 'in',
                ('para', 'cada'): 'for',  # token sequence translations
                ('faça', ':'): ':',
            }

    Now we can create an object with exec(), eval() and compile() functions that
    handle the newly defined transpyler:

        pybr = PyBr()
        global_ns = {}

        pybr.exec('''
        x, y = 1, 1
        para cada i em [1, 2, 3, 4, 5] faça:
            x, y = y, x + y
        ''', global_ns)

        assert globals_ns['x'] == 8
        assert globals_ns['y'] == 13
    """

    translations = None
    invalid_tokens = None
    language_version = '0.1.0'
    version = '0.1.0'
    codemirror_mode = 'python'
    file_extension = 'py'
    _compile = _compile
    _exec = _exec
    _eval = _eval
    _input = _input
    _print = _print
    lexer_factory = Lexer
    builtins_factory = Builtins
    info_factory = LanguageInfo

    @lazy
    def name(self):
        cls_name = self.__class__.__name__.lower()
        if cls_name == 'Transpyler':
            return 'transpyler'
        elif cls_name.endswith('transpyler'):
            return cls_name[:-10]
        else:
            return cls_name

    @lazy
    def mimetype(self):
        return 'text/x-%s' % self.name

    @lazy
    def display_name(self):
        return self.name.title()

    @lazy
    def lexer(self):
        return self.lexer_factory(self)

    @lazy
    def builtins(self):
        return self.builtins_factory(self)

    @lazy
    def info(self):
        return self.info_factory(self)

    @lazy
    def _namespace_cache(self):
        return self.get_builtins_namespace()

    def __init__(self, **kwargs):
        self._forbidden = False
        for k, v in kwargs.items():
            setattr(self, k, v)
        self._has_init = False

        # Check validity
        if self.name is None:
            raise ValueError('Transpyler requires a "name" attribute.')

        # Save instance in instance dictionary
        if self.name not in INSTANCES_FOR_NAME:
            INSTANCES_FOR_NAME[self.name] = self
        if type(self) not in INSTANCES_FOR_NAME:
            INSTANCES_FOR_NAME[type(self)] = self

    def __repr__(self):
        return '<%s: %r>' % (self.__class__.__name__, self.name)

    def update_user_ns(self, ns):
        """
        Update user namespace that is initialized when a Transpyled transpyler
        starts an interactive shell.
        """

    def init(self, extra_builtins=None, curses=True, apply_builtins=True):
        """
        Initializes transpyler runtime.

        Args:
            curses:
                Load curses that enable translation for builtin python types by
                hacking them at C-level.
            extra_builtins (dict):
                A dictionary with extra builtin functions to be added to the
                runtime.
        """

        if not self._has_init:
            if curses:
                self.apply_curses()
            self._has_init = True
        if extra_builtins:
            self.builtins.update(extra_builtins)
        if apply_builtins:
            self.builtins.load_as_builtins()

    def apply_curses(self):
        """
        Apply any required curses.

        Default implementation does nothing.
        """

    def compile(self, source, filename, mode, flags=0, dont_inherit=False,
                compile_function=None):
        """
        Similar to the built-in function compile() for Pytuguês code.

        The additional compile_function() argument allows to define a function to
        replace Python's builtin compile().

        Args:
            source (str or code):
                Code to be executed.
            filename:
                File name associated with code. Use '<input>' for strings.
            mode:
                One of 'exec' or 'eval'. The second allows only simple statements
                that generate a value and is used by the eval() function.
            forbidden (bool):
                If true, initialize the forbidden lib functionality to enable i18n
                for Python builtins in C-level.
            compile_function (callable):
                A possible replacement for Python's built-in compile().
        """

        source = self.transpile(source)
        compile_function = compile_function or self._compile
        return compile_function(source, filename, mode, flags, dont_inherit)

    def exec(self, source, globals=None, locals=None, forbidden=False,
             exec_function=None, builtins=False, _builtins_to_globals=True):
        """
        Similar to the built-in function exec() for Pytuguês code.

        The additional exec_function() argument allows to define a function to
        replace Python's builtin compile().

        Args:
            source (str or code):
                Code to be executed.
            globals, locals:
                A globals/locals dictionary
            builtins (bool):
                If given, execute with builtins injected on Python's builtins
                module.
            forbidden (bool):
                If true, initialize the forbidden lib functionality to enable i18n
                for Python builtins in C-level.
            exec_function (callable):
                A possible replacement for Python's built-in exec().
        """

        if builtins:
            with self.builtins.restore_after():
                return self.exec(source, globals, locals,
                                 forbidden=forbidden,
                                 _builtins_to_globals=False,
                                 exec_function=exec_function)

        exec_function = exec_function or _exec

        if globals is None:
            globals = _builtins.globals()
        if _builtins_to_globals:
            globals.update(self._namespace_cache)

        code = self.transpile(source) if isinstance(source, str) else source

        if locals is None:
            return exec_function(code, globals)
        else:
            return exec_function(code, globals, locals)

    def eval(self, source, globals=None, locals=None, forbidden=False,
             eval_function=None, builtins=False):
        """
        Similar to the built-in function eval() for transpyled code.

        The additional eval_function() argument allows to define a function to
        replace Python's builtin compile().

        Args:
            source (str or code):
                Code to be executed.
            globals, locals:
                A globals/locals dictionary
            builtins (bool):
                If given, update builtins with functions in tugalib.
            forbidden (bool):
                If true, initialize the forbidden lib functionality to enable
                i18n for Python builtins in C-level.
            eval_function (callable):
                A possible replacement for Python's built-in eval().
        """

        eval_function = eval_function or _eval
        return self.exec(source, globals, locals, forbidden=forbidden,
                         builtins=builtins, exec_function=eval_function)

    def transpile(self, src):
        """
        Convert source to Python.
        """

        return self.lexer.transpile(src)

    def is_incomplete_source(self, src, filename="<input>", symbol="single"):
        """
        Test if a given source code is incomplete.

        Incomplete code may appear in users interactions when user is typing a
        multi line command:

        for x in range(10):
            ... should continue here, but user already pressed enter!
        """

        pytuga_src = self.transpile(src)
        return codeop.compile_command(pytuga_src, filename, symbol) is None

    def get_builtins_namespace(self):
        """
        Return a dictionary with the default builtins get_namespace for transpyler.
        """

        return self.builtins.get_namespace()

    def get_console_banner(self):
        """
        Return a string with the console banner.
        """

        try:
            return self.banner
        except AttributeError:
            return 'Warning: please define .get_console_banner() method of your ' \
                   'transpyler.'


def get_transpyler_from_name(key):
    """
    Return a transpyler instance for the given transpyler name or type.
    """

    return INSTANCES_FOR_NAME[key]


# Defines the default transpyler instance
transpyler = Transpyler(name='default_transpyler')



from transpyler.utils import with_transpyler_attr


def run_jupyter(transpyler, gui=False):
    """
    Starts a Jupyter console for the given kernel.
    """

    from jupyter_client.kernelspec import NoSuchKernel
    from transpyler.jupyter.setup import setup_assets

    try:
        start_shell(transpyler, gui=gui)
    except NoSuchKernel:
        setup_assets(transpyler, user=True)
        start_shell(transpyler, gui=gui)


def start_console_shell(transpyler):
    """
    Starts a cli-based shell.
    """

    from transpyler.jupyter.app import ZMQTerminalTranspylerApp
    from transpyler.jupyter.app import TranspylerKernelManager

    App = with_transpyler_attr(ZMQTerminalTranspylerApp, transpyler)
    Manager = with_transpyler_attr(TranspylerKernelManager, transpyler)

    App.launch_instance(
        kernel_manager=Manager,
        kernel_name=transpyler.name,
        transpyler=transpyler,
    )


def start_qt_shell(transpyler):
    """
    Starts a shell based on QtConsole.
    """

    from transpyler.jupyter.app import ZMQTerminalTranspylerApp
    from transpyler.jupyter.app import TranspylerKernelManager

    App = with_transpyler_attr(ZMQTerminalTranspylerApp, transpyler)
    Manager = with_transpyler_attr(TranspylerKernelManager, transpyler)

    App.launch_instance(
        kernel_manager=Manager,
        kernel_name=transpyler.name,
        transpyler=transpyler,
    )


def start_shell(transpyler, gui=False):
    """
    Starts pytuga shell.
    """

    if gui:
        start_qt_shell(transpyler)
    else:
        start_console_shell(transpyler)



# -*- coding: utf-8 -*-
#
# This file were created by Python Boilerplate. Use boilerplate to start simple
# usable and best-practices compliant Python projects.
#
# Learn more about it at: http://github.com/fabiommendes/python-boilerplate/
#

import os
import codecs
from setuptools import setup, find_packages

# Save version and author to __meta__.py
version = open('VERSION').read().strip()
dirname = os.path.dirname(__file__)
path = os.path.join(dirname, 'src', 'transpyler', '__meta__.py')
meta = '''# Automatically created. Please do not edit.
__version__ = '%s'
__author__ = 'F\\xe1bio Mac\\xeado Mendes'
''' % version
with open(path, 'w') as F:
    F.write(meta)

setup(
    # Basic info
    name='transpyler',
    version=version,
    author='Fábio Macêdo Mendes',
    author_email='fabiomacedomendes@gmail.com',
    url='',
    description='A framework for building localized Python-like languages.',
    long_description=codecs.open('README.rst', 'rb', 'utf8').read(),

    # Classifiers (see https://pypi.python.org/pypi?%3Aaction=list_classifiers)
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
    ],

    # Packages and dependencies
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=[
        'lazyutils',
        'unidecode',
    ],
    extras_require={
        'dev': [
            'python-boilerplate[dev]',
        ],
        'google_translate': [
            'textblob',
        ],
        'jupyter': [
            'jupyter',
            'jupyter-console',
            'ipython',
        ],
    },

    # Other configurations
    zip_safe=False,
    platforms='any',
)


import sys

from mock import mock

from transpyler.tools.google_translate import translate_functions, \
    translate_function, get_args


def func(x, y=1, *args, z=1, d=2, **kwargs):
    """
    Some function!
    """


def test_get_correct_args():
    args = get_args(func)
    assert args == [('x', None), ('y', '1'), ('*args', None), ('z', '1'),
                    ('d', '2'), ('**kwargs', None)]


def test_google_translate_function():
    # Python 3.3 and 3.4 syntax breaks with complex g function.

    with mock.patch('textblob.TextBlob.translate', lambda x, y, z: str(x)):
        src = translate_function(func, 'pt-br')
    print(src)
    compile(src, '<input>', 'exec')


def test_google_translate_builtin():
    with mock.patch('textblob.TextBlob.translate', lambda x, y, z: str(x)):
        src = translate_functions([dir], 'pt-br')[0]
    compile(src, '<input>', 'exec')
